import scala.annotation.tailrec

@tailrec
def foldLeft[A, B](list: List[A])(init: B)(f: (B, A) => B): B = {
  list match {
    case Nil => init
    case head::tail => foldLeft(tail)(f(init, head))(f)
  }
}

@tailrec
def dropWhile[T](list: List[T])(predicate: T => Boolean): List[T] = {
  list match {
    case head::tail if predicate(head) => dropWhile(tail)(predicate)
    case _ => list
  }
}

def predicates[T](list: List[T])(preds: List[T => Boolean]): List[T] = {
  list.foldRight(List.empty[T])((e: T, l: List[T]) => {
    //if (preds.foldLeft(true)((a, b) => a && b(e))) e::l
    if (preds.forall(b => b(e))) e::l
    else l
  })
}

val l1 = (1 to 10).toList
val f1 = ((a: Int) => a % 2 == 0)
val f2 = ((a: Int) => a > 5)
predicates(l1)(List(f1, f2))  // List(6, 8, 10)

val l2 = ("Hello Scala Echo").toList
val f4 = ((a: Char) => a == 'a' || a == 'e' || a == 'o')
val f3 = ((a: Char) => !a.isUpper)
predicates(l2)(List(f3, f4))  // List('e', 'o', 'a', 'a', 'o')

/*
// Better solution
def predicates[T](list: List[T])(preds: List[T => Boolean]): List[T] = {
  preds.foldLeft(list)((acc, fun) => acc.filter(fun))
}
*/

def fixedPoint(f: Int => Int): Int => Int = {
  @tailrec
  def func(x: Int): Int = {
    val y: Int = f(x)
    if (x == y) y
    else func(y)
  }
  func
}

fixedPoint(x => if (x % 10 == 0) x else x + 1)(35)  // 40
fixedPoint(x => x / 2 + 5)(20)  // 10