import scala.annotation.tailrec

abstract class Text {
  def isEmpty: Boolean = {
    this match {
      case Chars(cs) => cs.isEmpty
      case Concat(t1, t2) => t1.isEmpty && t2.isEmpty
    }
  }
  def head: Char = {
    this match {
      case Chars(cs) => cs.head
      case Concat(t1, t2) => if (t1.isEmpty) t2.head else t1.head
    }
  }
  def tail: Text = {
    this match {
      case Chars(cs) => Chars(cs.tail)
      case Concat(t1, t2) => if (t1.isEmpty) t2.tail else Concat(t1.tail, t2)
    }
  }
  def map(f: Char => Char): Text = {
    this match {
      case Chars(cs) => Chars(cs.map(f))
      case Concat(t1, t2) => Concat(t1.map(f), t2.map(f))
    }
  }
}
case class Chars(cs: List[Char]) extends Text
case class Concat(t1: Text, t2: Text) extends Text

@tailrec
def equals(t1: Text, t2: Text): Boolean = {
  if (t1.isEmpty) t2.isEmpty
  else if (t2.isEmpty) false
  else (t1.head == t2.head) && equals(t1.tail, t2.tail)
}
